__title__ = "SSLyze"
__version__ = "5.0.0b0"
__description__ = "Fast and powerful SSL/TLS scanning library."
__author__ = "Alban Diquet"
__author_email__ = "nabla.c0d3@gmail.com"
__url__ = "https://github.com/nabla-c0d3/sslyze"
__license__ = "AGPL"
__copyright__ = "Copyright 2021 Alban Diquet"
